# -*- coding: utf-8 -*-
# (c) 2019 Andreas Motl <andreas@terkin.org>
# License: GNU Affero General Public License, Version 3
import logging
import sys

from munch import munchify

log = logging.getLogger(__name__)


def setup_logging(level=logging.INFO):
    log_format = "%(asctime)-15s [%(name)-22s] %(levelname)-7s: %(message)s"
    logging.basicConfig(format=log_format, stream=sys.stderr, level=level)


def configure_http_logging(options):
    # Control debug logging of HTTP requests.

    if options.http_logging:
        log_level = log.getEffectiveLevel()
    else:
        log_level = logging.WARNING

    requests_log = logging.getLogger("requests")
    requests_log.setLevel(log_level)

    requests_log = logging.getLogger("urllib3.connectionpool")
    requests_log.setLevel(log_level)


def normalize_options(options):
    normalized = {}
    for key, value in options.items():

        # Add primary variant.
        key = key.strip("--<>")
        normalized[key] = value

        # Add secondary variant.
        key = key.replace("-", "_")
        normalized[key] = value

    return munchify(normalized)



# -*- coding: utf-8 -*-
# (c) 2018-2021 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import logging

import jsonpickle
from docopt import docopt

from apicast import __appname__, __version__
from apicast.core import DwdBeeflightForecast
from apicast.format import Formatter
from apicast.util import normalize_options, setup_logging

log = logging.getLogger(__name__)


def run():
    """
    Access bee flight forecast information published by Deutscher Wetterdienst (DWD).

    Usage:
      apicast beeflight stations [--slugs]
      apicast beeflight forecast --station=<station> [--format=<format>]
      apicast service [--listen=<listen>] [--reload]
      apicast --version
      apicast (-h | --help)

    Options:
      --url=<url>                       URL to detail page
      --station=<station>               Station identifier
      --format=<format>                 Output format: "json", "json-machine" or "table-markdown". Default: json
      --listen=<listen>                 HTTP server listen address. [Default: localhost:24640]
      --version                         Show version information
      --debug                           Enable debug messages
      -h --help                         Show this screen

    Examples::

        # Display list of stations
        apicast beeflight stations

        # Display list of station slugs
        apicast beeflight stations --slugs

        # Display bee flight forecast for Potsdam in JSON format
        apicast beeflight forecast --station=brandenburg/potsdam

        # Display bee flight forecast for Potsdam in Markdown format
        apicast beeflight forecast --station=brandenburg/potsdam --format=table-markdown

        # Display bee flight forecast for Potsdam in JSON machine readable format
        apicast beeflight forecast --station=brandenburg/potsdam --format=json-machine

        # Start HTTP service
        apicast service

        # Start HTTP service with dynamic code reloading
        apicast service --reload

    """

    name = f"{__appname__} {__version__}"

    # Parse command line arguments
    options = normalize_options(docopt(run.__doc__, version=name))

    # Setup logging
    debug = options.get("debug")
    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    setup_logging(log_level)

    # Debugging
    log.debug("Options: {}".format(json.dumps(options, indent=4)))

    # Run service.
    if options.service:
        listen_address = options.listen
        log.info(f"Starting {name}")
        log.info(f"Starting web service on {listen_address}")
        from apicast.api import start_service

        start_service(listen_address, options.reload)
        return

    dbf = DwdBeeflightForecast()

    # Run command.
    if options.stations:

        if options.slugs:
            result = dbf.get_station_slugs()
            print("\n".join(result))

        else:
            result = dbf.get_stations()
            print(jsonpickle.encode(result, unpicklable=False, indent=4))

    # Fetch and extract forecast information.
    elif options.station:
        station = dbf.get_station_by_slug(options.station)
        result = dbf.get_data(station=station)
        format_beeflight_forecast(result, options.format)


def format_beeflight_forecast(result, format="json"):

    if not result.data:
        raise ValueError("No data found or unable to parse")

    format = format or "json"
    format = format.lower()

    if format == "json":
        response = Formatter(result).normalize()
        print(json.dumps(response, indent=4))

    elif format == "json-machine":
        response = Formatter(result).machinify()
        print(json.dumps(response, indent=4))

    elif format == "table-markdown":
        response = Formatter(result).table_markdown()
        print(response)

    else:
        raise ValueError(
            'Unknown output format. Please specify "json", "json-machine" or "table-markdown".'
        )



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import io
from contextlib import redirect_stdout
from datetime import datetime

import tabulate
from dateparser import parse as parsedate


class Formatter:
    """
    Die Bedingungen für den Bienenflug werden in 5 Intensitätsstufen angegeben
    (kein - gering - mittel - stark - intensiv).

    -- https://www.dwd.de/DE/fachnutzer/freizeitgaertner/dokumentation/gw_bienenflug
    """

    LABEL_MAP = {
        "Datum": "date",
        "morgens": "morning",
        "mittags": "noon",
        "abends": "evening",
        "kein": "no",
        "gering": "low",
        "mittel": "medium",
        "stark": "strong",
        "intensiv": "intensive",
    }

    STRENGTH_MACHINE_MAP = {
        "no": 0,
        "low": 1,
        "medium": 2,
        "strong": 3,
        "intensive": 4,
    }

    def __init__(self, result):
        self.result = result
        self.data = result.data
        self.title = u"### Prognose des Bienenfluges in {}".format(self.result.station_name)

    def translate(self):
        self.title = u"### Beeflight forecast for {}".format(self.result.station_name)
        for item in self.data:
            for index, slot in enumerate(item):
                for key, value in self.LABEL_MAP.items():
                    slot = slot.replace(key, value)
                item[index] = slot
        return self

    def normalize(self):
        result = []
        for item in self.data[1:]:
            item = dict(zip(self.data[0], item))
            result.append(item)
        return result

    def machinify(self):
        self.translate()
        data = self.normalize()
        for item in data:
            for key, value in item.items():
                if key == "date":
                    value += str(datetime.now().year)
                    value = parsedate(value).strftime("%Y-%m-%d")
                elif value in self.STRENGTH_MACHINE_MAP.keys():
                    value = self.STRENGTH_MACHINE_MAP[value]
                item[key] = value
        return data

    def table_markdown(self):
        with io.StringIO() as buffer, redirect_stdout(buffer):

            # Report about weather station / observation location
            print(self.title)
            print()

            # Output forecast data
            print(
                tabulate.tabulate(
                    self.data[1:],
                    headers=self.data[0],
                    showindex=False,
                    tablefmt="pipe",
                )
            )

            return buffer.getvalue()



# -*- coding: utf-8 -*-
# (c) 2020-2021 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging
from typing import List, Dict

from fastapi import FastAPI, Query
from fastapi.responses import HTMLResponse, PlainTextResponse

from apicast import __appname__, __version__
from apicast.core import (DwdBeeflightForecast, dwd_copyright, dwd_source,
                          producer_link, producer_name)
from apicast.format import Formatter

app = FastAPI()

log = logging.getLogger(__name__)

dbf = DwdBeeflightForecast()


@app.get("/", response_class=HTMLResponse)
def index():

    appname = f"{__appname__} {__version__}"
    description = "Apicast acquires bee flight forecast information published by Deutscher Wetterdienst (DWD)."

    data_index_items = []
    for location in dbf.get_station_slugs():
        item = f"""
            <li>
                <div style="float: left">
                Forecast for {location}
                </div>
                <div style="float: right">
                German:
                    <a href="beeflight/forecast/germany/{location}?translate=false&format=json">[JSON]</a>
                    <a href="beeflight/forecast/germany/{location}?translate=false&format=table-markdown">[Markdown]</a>
                English:
                    <a href="beeflight/forecast/germany/{location}?translate=true&format=json">[JSON]</a>
                    <a href="beeflight/forecast/germany/{location}?translate=true&format=table-markdown">[Markdown]</a>
                Machine:
                    <a href="beeflight/forecast/germany/{location}?format=json-machine">[JSON]</a>
                </div>
                <div style="clear: both"/>
            </li>
        """
        data_index_items.append(item)

    data_index_items_html = "\n".join(data_index_items)

    return f"""
    <html>
        <head>
            <title>{appname}</title>

            <meta name="description"
                content="{description}">
            <meta name="keywords"
                content="honey bee, apis mellifera, flight forecast information, dwd, cdc, deutscher wetterdienst, climate data center, weather, opendata, data acquisition, transformation, export, geospatial, temporal, timeseries, sensor, network, observation, http, rest, api, json, markdown">

            <meta property="og:title" content="{appname}"/>
            <meta property="og:site_name" content="{appname}">
            <meta property="og:description"
                content="{description}"/>
            <meta property="og:url" content="https://apicast.hiveeyes.org/"/>
            <meta property="og:image" content="https://ptrace.hiveeyes.org/2021-03-05_beeswarm_plymouth.jpg"/>
            <meta property="og:image:secure_url" content="https://ptrace.hiveeyes.org/2021-03-05_beeswarm_plymouth.jpg"/>
            <meta property="og:type" content="website"/>

            <meta name="twitter:title" content="{appname}">
            <meta name="twitter:description" content="{description}">
            <meta name="twitter:image" content="https://ptrace.hiveeyes.org/2021-03-05_beeswarm_plymouth.jpg">
            <meta name="twitter:image:alt" content="Bee swarm in Plymouth">
            <meta name="twitter:card" content="summary_large_image">

            <script type='application/ld+json'>{{"@context":"https://schema.org","@type":"WebSite","@id":"#website","url":"https://apicast.hiveeyes.org/","name":"{appname}","logo":"https://ptrace.hiveeyes.org/2021-03-05_beeswarm_plymouth.jpg"}}</script>

        </head>
        <body>
            <h2>About</h2>
            <div style="width: 68%">
                <div style="float: left">
                {description}
                <ul>
                <li>Producer: <a href="{producer_link}">{producer_name}</a></li>
                <li>Source: <a href="{dwd_source}">DWD » Freizeitgärtner » Gartenwetter » Prognose des Bienenfluges</a></li>
                <li>Data copyright: {dwd_copyright}</li>
                </ul>
                </div>
                <div style="float: right">
                    <img src="https://ptrace.hiveeyes.org/2021-03-05_beeswarm_plymouth.jpg" width="200"/>
                </div>
            </div>
            <div style="clear: both"/>

            <h3>Location index</h3>
            <ul>
            <li><a href="beeflight/stations/germany">List of federal states / stations</a></li>
            <li><a href="beeflight/stations/germany/locations">List of location slugs</a></li>
            </ul>
            <h3>Data index</h3>
            <ul style="width: 65%">
            {data_index_items_html}
            </ul>
        </body>
    </html>
    """


@app.get("/robots.txt", response_class=PlainTextResponse)
def robots():
    return f"""
User-agent: *
Disallow: /beeflight/
    """.strip()


@app.get("/beeflight/stations/germany")
def beeflight_stations():
    stations = dbf.get_stations()
    return make_json_response(data=stations, location="germany")


@app.get("/beeflight/stations/germany/locations")
def beeflight_stations_site_slugs():
    slugs = dbf.get_station_slugs()
    return make_json_response(data=slugs, location="germany")


@app.get("/beeflight/forecast/germany/{state}/{station}")
def beeflight_forecast_by_slug(
    state: str,
    station: str,
    format: str = Query(default="json"),
    translate: bool = Query(default=False),
):

    station_slug = f"{state}/{station}"

    try:
        station = dbf.get_station_by_slug(station_slug)
        result = dbf.get_data(station=station).copy()
        if not result.data:
            raise ValueError("No data found or unable to parse")
    except Exception as ex:
        return {"error": str(ex)}

    formatter = Formatter(result)

    if translate:
        formatter.translate()

    if format == "json-machine":
        response = formatter.machinify()

    elif format == "table-markdown":
        return PlainTextResponse(formatter.table_markdown())

    else:
        response = formatter.normalize()

    return make_json_response(data=response, location=station_slug)


def make_json_response(data: List[Dict], location: str = None):
    response = {
        "meta": {
            "source": dwd_source,
            "producer": f"{producer_name} - {producer_link}",
            "copyright": dwd_copyright,
        },
        "location": {
            "slug": location,
        },
        "data": data,
    }
    return response


def start_service(listen_address, reload: bool = False):
    host, port = listen_address.split(":")
    port = int(port)
    from uvicorn.main import run

    run(app="apicast.api:app", host=host, port=port, reload=reload)



# -*- coding: utf-8 -*-
# (c) 2018-2021 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
"""
Apicast acquires bee flight forecast information published by Deutscher Wetterdienst (DWD).

Parse information about "Gartenwetter » Bienenflug"
from https://www.dwd.de/DE/leistungen/biene_flug/bienenflug.html.

See also https://community.hiveeyes.org/t/dwd-prognose-bienenflug/787
"""
import dataclasses
from typing import List

import requests
import ttl_cache
from bs4 import BeautifulSoup
from html_table_extractor.extractor import Extractor
from slugify import slugify

from apicast import __appname__, __version__


user_agent = f"{__appname__}/{__version__}"
dwd_source = "https://www.dwd.de/DE/leistungen/biene_flug/bienenflug.html"
dwd_copyright = "© Deutscher Wetterdienst (DWD), Agricultural Meteorology Department"
producer_name = "Hiveeyes Apicast"
producer_link = "https://github.com/hiveeyes/apicast"


@dataclasses.dataclass
class State:
    label: str
    identifier: str


@dataclasses.dataclass
class Station:
    state: State
    label: str
    identifier: str
    slug: str


@dataclasses.dataclass
class Result:
    station: Station
    station_name: str
    data: List
    footnote: str

    def copy(self):
        return Result(**dataclasses.asdict(self))


class DwdBeeflightForecast:

    baseurl = "https://www.dwd.de/DE/leistungen/biene_flug/bienenflug.json?cl2Categories_LeistungsId=bienenflug"

    session = requests.Session()
    session.headers["User-Agent"] = user_agent

    @ttl_cache(60 * 60 * 24)
    def get_states(self) -> List[State]:

        states: List[State] = []

        # Request federal states.
        response = self.session.get(
            self.baseurl, params={"view": "renderJson", "cl2Categories_Bundesland": ""}
        )
        data = response.json()
        for item in data["cl2Categories_Bundesland"].values():
            state = State(label=item["label"], identifier=item["val"])

            states.append(state)

        # sites.sort(key=operator.attrgetter("label"))

        return states

    @ttl_cache(60 * 60 * 24)
    def get_stations(self) -> List[Station]:

        stations: List[Station] = []

        # Request federal states.
        for state in self.get_states():

            # Request sites.
            response = self.session.get(
                self.baseurl,
                params={
                    "view": "renderJson",
                    "cl2Categories_Bundesland": state.identifier,
                    "cl2Categories_Station": "",
                },
            )
            data = response.json()
            for item in data["cl2Categories_Station"].values():
                identifier = item["val"]
                if identifier == "bifl_0000":
                    continue
                label = item["label"]
                station = Station(
                    state=state,
                    label=label,
                    identifier=identifier,
                    slug=f"{slugify(state.label)}/{slugify(label)}",
                )

                stations.append(station)

        # sites.sort(key=operator.attrgetter("label"))

        return stations

    def get_station_slugs(self) -> List[str]:
        slugs = []
        for station in self.get_stations():
            slugs.append(station.slug)
        slugs.sort()
        return slugs

    def get_station_by_slug(self, slug):
        for station in self.get_stations():
            if station.slug == slug:
                return station
        raise KeyError("No such station")

    @ttl_cache(60 * 60)
    def get_data(self, station: Station) -> Result:

        response = self.session.get(
            self.baseurl,
            params={
                "view": "renderJsonResults",
                "cl2Categories_Bundesland": station.state.identifier,
                "cl2Categories_Station": station.identifier,
            },
        )

        soup = BeautifulSoup(markup=response.content, features="html.parser")

        # Find content section and extract elements.
        headline = soup.find(string="Prognose der Bienenflugintensität")
        if not headline:
            raise ValueError("No forecast available for this station")
        station_name = headline.find_next("section").attrs.get("aria-label")
        table = soup.find("table")

        footnote = soup.find("aside").find("p").text

        # Read HTML table
        data = self.parse_html_table(table)

        # Ready.
        result = Result(
            station=station, station_name=station_name, data=data, footnote=footnote
        )
        return result

    @staticmethod
    def parse_html_table(html):
        extractor = Extractor(html)
        extractor.parse()
        return extractor.return_list()



import dataclasses

from datadiff.tools import assert_equal

from apicast.core import DwdBeeflightForecast

dbf = DwdBeeflightForecast()


def test_get_states():

    states = dbf.get_states()
    bavaria = [state for state in states if state.identifier == "bifl_bl02"][0]

    assert_equal(
        dataclasses.asdict(bavaria),
        {
            "label": "Bayern",
            "identifier": "bifl_bl02",
        },
    )


def test_get_stations():

    stations = dbf.get_stations()
    bavaria_hof = [
        station for station in stations if station.identifier == "bifl_0042"
    ][0]
    berlin_tempelhof = [
        station for station in stations if station.identifier == "bifl_0019"
    ][0]

    assert_equal(
        dataclasses.asdict(bavaria_hof),
        {
            "label": "Hof",
            "identifier": "bifl_0042",
            "slug": "bayern/hof",
            "state": {
                "label": "Bayern",
                "identifier": "bifl_bl02",
            },
        },
    )

    assert_equal(
        dataclasses.asdict(berlin_tempelhof),
        {
            "label": "Berlin-Tempelhof",
            "identifier": "bifl_0019",
            "slug": "berlin/berlin-tempelhof",
            "state": {
                "label": "Berlin",
                "identifier": "bifl_bl03",
            },
        },
    )



import dataclasses

import pytest
from datadiff.tools import assert_equal

from apicast.core import DwdBeeflightForecast

dbf = DwdBeeflightForecast()


@pytest.mark.data
def test_get_data():

    stations = dbf.get_stations()
    bavaria_hof = [
        station for station in stations if station.identifier == "bifl_0042"
    ][0]

    data_hof = dbf.get_data(station=bavaria_hof)

    data = data_hof.data
    assert_equal(data[0], ["Datum", "morgens", "mittags", "abends"])

    footnote = data_hof.footnote
    assert "© Deutscher Wetterdienst, erstellt" in footnote
    assert "Alle Angaben ohne Gewähr!" in footnote

    data_hof.data = None
    data_hof.footnote = None

    assert_equal(
        dataclasses.asdict(data_hof),
        {
            "station": {
                "state": {"label": "Bayern", "identifier": "bifl_bl02"},
                "label": "Hof",
                "identifier": "bifl_0042",
                "slug": "bayern/hof",
            },
            "station_name": "Hof",
            "data": None,
            "footnote": None,
        },
    )
